package imp.slicer;

import java.util.Vector;

import imp.parser.Loc;
import imp.util.VarIdSet;

public abstract class Node {
    public int indentLevel;				// for formating 
    public Loc loc;						// location of the node
    public Vector prevs; 				// predecessors
    public Vector nexts;				// successors; for IMP at most two
    public boolean isRelevant;			// true iff this node is relevant		
    public VarIdSet dRVars;				// set of directly relevant variables
    public boolean dRVarsChanged;		// needed for implementing slicing on programs
    									// with loops
    
    public Node (int i, Loc l) {
    	this.indentLevel = i;
    	this.loc = l;
    	this.prevs = new Vector();
    	this.nexts = new Vector();
    	this.isRelevant = false;		// by default, no node is relevant
    	this.dRVars = new VarIdSet();
    	this.dRVarsChanged = false;
    }
    
    /**
     * Add the argument node to the vector of predecessors
     */
    public boolean addPrev(Node n) {
      return prevs.add(n);
    }
    
    /**
     * Add the argument node to the vector of successors
     */
    public boolean addNext(Node n) {
      return nexts.add(n);
    }
    
    /**
     * Assumes that the node has only one successor. Returns that node.
     * TestNodes (IfThenTestNode, IfThenElseTestNode,
     * WhileTestNode, RepeatTestNode) have two successors.
     * ProgramEndNode has no successor. All other nodes (AssignNode, SkipNode,
     * PrintNode, TestEndNode) have one successor.
     * So, getNext() should only get invoked on these four nodes.
     */
    public Node getNext() {
    	if (nexts.size() != 1) {
    		System.out.println("Successor node not unique were it should be\n");
    		System.out.println(this.toString());
    		System.out.println(nexts.size());
    		System.exit(1);
    	}
    	return (Node) nexts.elementAt(0);
    }
    

    /** 
     * Creates string representation of this node if it is relevant (by looking
     * at the "isRelevant" flag). Otherwise, it returns "". 
     * Assumes that this node has exactly one successor. 
     * Overwritten for all objects that don't:
     * IfThenTestNode, IfThenElseTestNode, WhileTestNode, RepeatTestNode.
     */
    public String toStringUsingRelevant(Node stopNode) {
    	if (this != stopNode) {
    		String str = this.toString();
    		Node nextNode = this.getNext();
    		if (this instanceof ProgramBeginNode || this instanceof ProgramEndNode) 
    			// ProgramBeginNode and ProgramEndNods are always relevant
    			return this.toString() + "\n" + (nextNode.toStringUsingRelevant(stopNode));
    		else if (this.isRelevant) {
    			// make sure we don't get ";" before an "END" or an "UNTIL"
        		if (this instanceof RepeatNode || 
        				nextNode instanceof EndNode || 
        				nextNode instanceof RepeatTestNode ||
        				(nextNode == stopNode && nextNode instanceof WhileTestNode))
        			return str + "\n" + (nextNode.toStringUsingRelevant(stopNode));
        		else
        			return str + ";\n" + (nextNode.toStringUsingRelevant(stopNode));	
    		}
    		else 
    			return (nextNode.toStringUsingRelevant(stopNode));
    	}
    	else 
    		return "";
  	}

    /**
     * Creates string representation of this node
     */
    public abstract String toString();
    
    /**
     * Creates the string representation of all nodes between this node and 
     * the node before the stopNode
     */
    public String toString(Node stopNode) {
      	if (this != stopNode) {
      		String str = this.toString();
      		Node nextNode = this.getNext();
      		// make sure we don't get ";" before an "END" or an "UNTIL"
      		if ((nextNode instanceof EndNode) || 
      			(nextNode instanceof RepeatTestNode) ||
      			(this instanceof ProgramBeginNode))
    			return str + "\n" + (nextNode.toString(stopNode));
    		else
    			return str + ";\n" + (nextNode.toString(stopNode));
      	}
      	else 
      		return "";
      }
    
    /**
     * Computes directly relevant variables in this node and all its 
     * predecessors
     */
    public abstract void computeDRVars(Node cNode, VarIdSet cVars);
    
} // public class Node
